/*************************************************************************
|  COPYRIGHT (c) 2000 BY ABATRON AG
|*************************************************************************
|
|  PROJECT NAME: BDI Communication Driver
|  FILENAME    : bdilnk.c
|
|  COMPILER    : GCC
|
|  TARGET OS   : LINUX
|  TARGET HW   : PC
|
|  PROGRAMMER  : Abatron / RD
|  CREATION    : 19.06.00
|
|*************************************************************************
|
|  DESCRIPTION :
|  Data link functions for communication with BDI
|
|
|*************************************************************************/

/*************************************************************************
|  INCLUDES
|*************************************************************************/

#include <ctype.h>
#include <unistd.h>
#include <errno.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>

#include <sys/types.h>
#include <sys/time.h>
#include <sys/times.h>
#include <sys/socket.h>
#include <sys/select.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include "compress.h"
#include "bdierror.h"
#include "bdicmd.h"
#include "bdilnkmt.h"

/*************************************************************************
|  DEFINES
|*************************************************************************/

#define DLE     16
#define STX      2
#define ETX      3

#define NET_TRANSFER_TIMEOUT            100

/* Frame control */
#define FRAME_COUNT_FIELD               (3<<6)
#define FRAME_LENGTH_MASK               7
#define FRAME_TYPE_MASK                 (7<<3)
#define FRAME_LNK_TYPE                  (0<<3)
#define FRAME_ATT_TYPE                  (1<<3)
#define FRAME_STD_TYPE                  (2<<3)
#define FRAME_NET_TYPE                  (6<<3)

/* link management commands */
#define LNK_RESET                       1
#define LNK_ECHO                        2
#define LNK_SET_BAUDRATE                3


/*************************************************************************
|  MACROS
|*************************************************************************/

/*************************************************************************
|  TYPEDEFS
|*************************************************************************/

/*************************************************************************
|  LOCALS
|*************************************************************************/

/****************************************************************************
 ****************************************************************************

    BDI_Append___ :

    Host independent helper function to append a numeric value to a buffer.
    The bytes will be stored in the link format order (Motorola byte order).

     INPUT  : value     value
              buffer    pointer to buffer
     OUTPUT : RETURN    pointer to next byte after the stored value
     INOUT  :

 ****************************************************************************/

static BYTE* BDI_AppendByte(BYTE  value, BYTE* buffer)
{
  *buffer++ = value;
  return buffer;
} /* BDI_AppendByte */

static BYTE* BDI_AppendWord(WORD value, BYTE* buffer)
{
  *buffer++ = (BYTE)(value>>8);
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendWord */

static BYTE* BDI_AppendLong(DWORD value, BYTE* buffer)
{
  *buffer++ = (BYTE)(value>>24);
  *buffer++ = (BYTE)(value>>16);
  *buffer++ = (BYTE)(value>>8);
  *buffer++ = (BYTE)value;
  return buffer;
} /* BDI_AppendLong */


/****************************************************************************
 ****************************************************************************

    BDI_Extract___ :

    Host independent helper function to extract a numeric value from a buffer.
    The value must be stored in link format (Motorola byte order).

     INPUT  : buffer    pointer to buffer
     OUTPUT : value     pointer to variable where to store the value
              RETURN    pointer to next byte after the extracted value
     INOUT  :

 ****************************************************************************/

static BYTE* BDI_ExtractByte(BYTE* value, BYTE* buffer)
{
  *value = *buffer++;
  return buffer;
} /* BDI_ExtractByte */

static BYTE* BDI_ExtractWord(WORD* value, BYTE* buffer)
{
  WORD x;
  x = (WORD)*buffer++;
  x = (x<<8) + (WORD)*buffer++;
  *value = x;
  return buffer;
} /* BDI_ExtractWord */

static BYTE* BDI_ExtractLong(DWORD* value, BYTE* buffer)
{
  DWORD x;
  x = (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  x = (x<<8) + (DWORD)*buffer++;
  *value = x;
  return buffer;
} /* BDI_ExtractLong */


/****************************************************************************
 ****************************************************************************
    BDI_DoDelay
    Helper function to delay an amount of time

     INPUT:  delay      the delay time in ms
     OUTPUT: -
 ****************************************************************************/

static void BDI_DoDelay(DWORD delay)
{
  usleep(1000 * delay);
} /* BDI_DoDelay */


/****************************************************************************
 ****************************************************************************
                Asynchronous Communication Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
    SetBaudrate

     INPUT:  baudrate       sets the baudrate of the communication port
     OUTPUT: return         error code
 ****************************************************************************/

static int AsynSetBaudrate(BDI_ChannelT* pCh, DWORD baudrate)
{
  struct termios        tios;
  speed_t               speed;

  /* get terminal attributes */
  if (tcgetattr(pCh->fd, &tios) < 0) return BDI_ASYN_SETUP;

  /* set new baudrate */
  pCh->asynBaudrate = baudrate;
  if      (baudrate ==   9600) speed = B9600;
  else if (baudrate ==  19200) speed = B19200;
  else if (baudrate ==  38400) speed = B38400;
  else if (baudrate ==  57600) speed = B57600;
  else if (baudrate == 115200) speed = B115200;
  else                         speed = B9600;
  cfsetospeed(&tios, speed);
  cfsetispeed(&tios, speed);

  /* set terminal attributes */
  if (tcsetattr(pCh->fd, TCSAFLUSH, &tios) < 0) return BDI_ASYN_SETUP;

  return BDI_OKAY;
} /* AsynSetBaudrate */


/****************************************************************************
    AsynOpen
    Opens the asynchronous communication channel

     INPUT:  port           the communication port e.g. "COM1"
     OUTPUT: return         error code
 ****************************************************************************/

static int AsynOpen(BDI_ChannelT* pCh, const char* port)
{
  int                   fd;
  struct termios        tios;

  /* open device */
  fd = open(port, O_RDWR | O_NONBLOCK | O_NOCTTY);
  if (fd < 0) return BDI_ASYN_SETUP;
  pCh->fd = fd;

  /* get terminal attributes */
  if (tcgetattr(fd, &tios) < 0) return BDI_ASYN_SETUP;

  /* set terminal attributes */
  pCh->asynBaudrate = 9600;
  tios.c_iflag = 0;
  tios.c_oflag = 0;
  tios.c_cflag = CLOCAL | CREAD | CS8;
  tios.c_lflag = 0;
  tios.c_cc[VMIN]  = 0;
  tios.c_cc[VTIME] = 1;
  cfsetospeed(&tios, B9600);
  cfsetispeed(&tios, B9600);
  if (tcsetattr(fd, TCSAFLUSH, &tios) < 0) return BDI_ASYN_SETUP;

  return BDI_OKAY;
} /* AsynOpen */


/****************************************************************************
    Closes the communication port

     INPUT:  -
     OUTPUT:
 ****************************************************************************/

static void AsynClose(BDI_ChannelT* pCh)
{
  close(pCh->fd);
} /* AsynClose */


/****************************************************************************
    Checks if characters are in the receive queue

     INPUT:  -
     OUTPUT: return     number of characters in the receive queue
 ****************************************************************************/

static int AsynInRxQueue(BDI_ChannelT* pCh)
{
  return 0;
} /* AsynInRxQueue */


/****************************************************************************
    Reads a byte from the communication port

     INPUT:  timeout        timeout in milliseconds
     OUTPUT: c              received byte
             return         0 = okay, else error
 ****************************************************************************/

static int AsynReadCharB(BDI_ChannelT* pCh, BYTE *c, DWORD timeout)
{
  int            rxCount;
  fd_set         readset;
  struct timeval rxTimeout;

  rxTimeout.tv_sec  = timeout / 1000;
  rxTimeout.tv_usec = (timeout % 1000) * 1000;
  FD_ZERO(&readset);
  FD_SET(pCh->fd, &readset);
  rxCount = select(pCh->fd + 1, &readset, NULL, NULL, &rxTimeout);
  if (rxCount > 0) {
    rxCount = read(pCh->fd, c, 1);
    if (rxCount ==  1) return BDI_OKAY;
    else               return BDI_ASYN_RX_ERROR;
  } /* if */
  else if (rxCount == 0) {
    return BDI_ASYN_RX_TIMEOUT;
  } /* else if */
  else {
    return BDI_ASYN_RX_ERROR;
  } /* else */
} /* NetWaitFrame */

static int AsynReadChar(BDI_ChannelT* pCh, BYTE *c, DWORD timeout)
{
  struct tms    tbuf;
  DWORD         startTime;

  if (read(pCh->fd, c, 1) == 1) return BDI_OKAY; /* high speed exit */
  timeout = (timeout * sysconf(_SC_CLK_TCK)) / 1000L;
  startTime = times(&tbuf);
  for (;;) {
    if ((times(&tbuf) - startTime) > timeout) return BDI_ASYN_RX_TIMEOUT;
    if (read(pCh->fd, c, 1) == 1)      return BDI_OKAY;
  } /* for */
} /* AsynReadChar */


/****************************************************************************
    Writes a block to the communication port

     INPUT:  count          number of bytes to send
             data           the data to send
     OUTPUT: return         0 = okay, else error

 ****************************************************************************/

static int AsynWriteBlock(BDI_ChannelT* pCh, DWORD count, BYTE* data)
{
  int   written;

  while (count > 0) {
    written = write(pCh->fd, data, count);
    if (written > 0) {
      count -= (DWORD)written;
      data  += written;
    } /* if */
    else if ((written == -1) && (errno == EAGAIN)) {
      continue;
    } /* else if */
    else if (written < 0) {
      return BDI_ASYN_TX_ERROR;
    } /* else */
  } /* while */
  return BDI_OKAY;
} /* AsynWriteBlock */


/****************************************************************************
    Sends a BDI frame

     INPUT:  count          number of bytes to send
             frame          the frame to send
     OUTPUT: return         0 = okay, else error

     OUTPUT:
 ****************************************************************************/

static int AsynSendFrame(BDI_ChannelT* pCh, int count, BYTE* frame)
{
  BYTE  bcc;
  BYTE  txChar;
  BYTE* buffer;
  BYTE  asynTxBuffer[BDI_MAX_FRAME_SIZE];

  buffer = asynTxBuffer;

  /* send start sequence */
  *buffer++ = DLE;
  *buffer++ = STX;

  /* send frame data */
  bcc = 0;
  while (count--) {
    txChar    = *frame++;
    *buffer++ = txChar;
    bcc      ^= txChar;
    if (txChar == DLE) *buffer++ = DLE;
  } /* while */

  /* send end sequence */
  *buffer++ = DLE;
  *buffer++ = ETX;
  *buffer++ = bcc;
  if (bcc == DLE) *buffer++ = DLE;

  /* send prepared frame */
  return AsynWriteBlock(pCh, buffer - asynTxBuffer, asynTxBuffer);
} /* AsynSendFrame */


/****************************************************************************
    Gets a BDI frame

     INPUT:  count          the maximal number of byte to receive
             timeout        the maximal time to wait for the frame in ms
     OUTPUT: frame          the received frame
             return         the size of the received frame or error

     OUTPUT:
 ****************************************************************************/

static int AsynWaitFrame(BDI_ChannelT* pCh, int count, BYTE* frame, DWORD timeout)
{
  BYTE  bcc;
  BYTE* putPtr;
  int   result;
  int   rxCount;
  BYTE  rxChar;

  /* wait for start sequence */
  for (;;) {
    result = AsynReadChar(pCh, &rxChar, timeout);
    if (result == BDI_ASYN_RX_TIMEOUT) return BDI_ASYN_RX_TIMEOUT;
    if ((result == BDI_OKAY) && (rxChar == DLE)) {
      result = AsynReadChar(pCh, &rxChar, timeout);
      if (result == BDI_ASYN_RX_TIMEOUT) return BDI_ASYN_RX_TIMEOUT;
      if ((result == BDI_OKAY) && (rxChar == STX)) break;
    } /* if */
  } /* for */

  /* receive frame data */
  bcc        = 0;
  rxCount    = 0;
  putPtr     = frame;
  for (;;) {

    /* get next char */
    result = AsynReadChar(pCh, &rxChar, timeout);
    if (result != BDI_OKAY) return result;

    /* process link escape char */
    if (rxChar == DLE) {
      result = AsynReadChar(pCh, &rxChar, timeout);
      if (result != BDI_OKAY) return result;

      /* process end sequence */
      if (rxChar == ETX) {
        result = AsynReadChar(pCh, &rxChar, timeout);
        if (result != BDI_OKAY) return result;
        if (rxChar == DLE) {
          result = AsynReadChar(pCh, &rxChar, timeout);
          if (result != BDI_OKAY) return result;
          if (rxChar != DLE)      return BDI_ASYN_RX_FORMAT;
        } /* if */
        if (rxChar == bcc) return rxCount;
        else               return BDI_ASYN_RX_BCC;
      } /* if */
      else if (rxChar != DLE) return BDI_ASYN_RX_FORMAT;
    } /* if */

    /* store data in receive buffer */
    if (rxCount < count) {
      bcc      ^= rxChar;
      *putPtr++ = rxChar;
      rxCount++;
    } /* if */
    else return BDI_ASYN_RX_OVERFLOW;
  } /* for */

} /* AsynWaitFrame */


/****************************************************************************
    Reset Link
    Search the current baudrate, reset the link and set new baudrate.

     INPUT:  baudrate       the requested baudrate
     OUTPUT: return         error code

     OUTPUT:
 ****************************************************************************/

#define NBR_OF_BAUDRATES     5

static int AsynResetLink(BDI_ChannelT* pCh, DWORD baudrate)
{
  static const DWORD rateTable[NBR_OF_BAUDRATES] = {9600,19200,38400,57600,115200};

  int   topRate;
  int   rate;
  int   rxCount;
  int   txCount;
  int   result;
  BYTE* txPtr;
  BYTE* rxPtr;
  DWORD confirmedRate;

  /* check if 115200 is supported */
  topRate = NBR_OF_BAUDRATES - 1;
  if (AsynSetBaudrate(pCh, rateTable[topRate]) != BDI_OKAY) topRate--;

  /* adjust requested rate if too high */
  if (baudrate > rateTable[topRate]) baudrate = rateTable[topRate];

  /* try to connect with requested baudrate */
  txPtr    = pCh->txFrame;
  *txPtr++ = FRAME_LNK_TYPE;
  *txPtr++ = 1;
  *txPtr++ = LNK_RESET;
  txCount  = txPtr - pCh->txFrame;
  rxCount  = 0;
  result = AsynSetBaudrate(pCh, baudrate);
  if (result == BDI_OKAY) result  = AsynSendFrame(pCh, txCount, pCh->txFrame);
  if (result == BDI_OKAY) rxCount = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 200);
  if (rxCount == txCount) return BDI_OKAY;

  /* try all baudrates */
  for (rate=0; rate<=topRate; rate++) {
    result = AsynSetBaudrate(pCh, rateTable[rate]);
    if (result == BDI_OKAY) result  = AsynSendFrame(pCh, txCount, pCh->txFrame);
    if (result == BDI_OKAY) rxCount = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 200);
    if (rxCount == txCount) break;
  } /* for */
  if (rxCount != txCount) return BDI_ERR_NO_RESPONSE;

  /* change baudrate */
  txPtr    = pCh->txFrame;
  *txPtr++ = FRAME_LNK_TYPE;
  *txPtr++ = 5;
  *txPtr++ = LNK_SET_BAUDRATE;
  txPtr    = BDI_AppendLong(baudrate, txPtr);
  txCount  = txPtr - pCh->txFrame;
  rxCount  = 0;
  result = AsynSendFrame(pCh, txCount, pCh->txFrame);
  if (result == BDI_OKAY) rxCount = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 200);
  if (rxCount != txCount) return BDI_ASYN_SETUP;

  /* give BDI time to change baudrate */
  BDI_DoDelay(300);

  /* extract confirmed baudrate */
  rxPtr = BDI_ExtractLong(&confirmedRate, pCh->rxFrame+3);

  /* check if success */
  txPtr    = pCh->txFrame;
  *txPtr++ = FRAME_LNK_TYPE;
  *txPtr++ = 1;
  *txPtr++ = LNK_RESET;
  txCount  = txPtr - pCh->txFrame;
  rxCount  = 0;
  result = AsynSetBaudrate(pCh, confirmedRate);
  if (result == BDI_OKAY) result  = AsynSendFrame(pCh, txCount, pCh->txFrame);
  if (result == BDI_OKAY) rxCount = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 200);

  /* return result */
  if      (rxCount == txCount) return BDI_OKAY;
  else if (result != BDI_OKAY) return result;
  else if (rxCount < 0)        return rxCount;
  else                         return BDI_ERR_INVALID_RESPONSE;
} /* AsynResetLink */


/****************************************************************************
 ****************************************************************************
                Network Communication Functions
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
    NetOpen
    Opens a TCP/IP socket

     INPUT:  szAddr         the ip address as string
             port           the ip port
     OUTPUT: return         error code
 ****************************************************************************/

static int NetOpen(BDI_ChannelT* pCh, const char* szAddr, WORD port)
{
  struct    sockaddr_in   bdiNetAddr;
  int       err;
  DWORD     ipAddr;
  struct    hostent *pHost;  /* pointer to host info for remote host */


  /* first check to see if the host name was passed in as dot notation */
  /* A return value of INADDR_NONE means inet_addr() failed, so try gethostbyname() */
  ipAddr = inet_addr(szAddr);
  if (ipAddr == INADDR_NONE) {
    pHost = gethostbyname(szAddr);
    if (pHost != NULL) {
      ipAddr = ((struct in_addr *)(pHost->h_addr))->s_addr;
    } /* if */
  } /* if */

  /* return with error if no host information */
  if (ipAddr == INADDR_NONE) {
    return BDI_ERR_INVALID_PARAMETER;
  } /* if */

  /* open socket */
  pCh->fd = socket(AF_INET, SOCK_DGRAM, 0);
  if (pCh->fd == -1) {
    return BDI_SOCKET_ERROR;
  } /* if */

  /* open connection */
  bdiNetAddr.sin_family      = AF_INET;
  bdiNetAddr.sin_port        = htons(port);
  bdiNetAddr.sin_addr.s_addr = ipAddr;
  err = connect(pCh->fd, (struct sockaddr*)&bdiNetAddr, sizeof bdiNetAddr);
  if (err == -1) {
    err = close(pCh->fd);
    pCh->fd = -1;
    return BDI_SOCKET_ERROR;
  }

  return BDI_OKAY;
} /* NetOpen */

/****************************************************************************
    Closes the TCP/IP socket

     INPUT:  -
     OUTPUT:
 ****************************************************************************/

static void NetClose(BDI_ChannelT* pCh)
{
  int       err;

  if (pCh->fd != -1) {
    err = close(pCh->fd);
  } /* if */
} /* NetClose */


/****************************************************************************
    Sends a BDI frame

     INPUT:  count          number of bytes to send
             frame          the frame to send
     OUTPUT: return         0 = okay, else error

     OUTPUT:
 ****************************************************************************/

static int NetSendFrame(BDI_ChannelT* pCh, int count, BYTE* frame)
{
  int result;

  result = send(pCh->fd, frame, count, 0);
  if (result != count) return BDI_SOCKET_ERROR;
  else                 return BDI_OKAY;
} /* NetSendFrame */


/****************************************************************************
    Gets a BDI frame

     INPUT:  count          the maximal number of byte to receive
             timeout        the maximal time to wait for the frame in ms
     OUTPUT: frame          the received frame
             return         the size of the received frame or error

     OUTPUT:
 ****************************************************************************/

static int NetWaitFrame(BDI_ChannelT* pCh, int count, BYTE* frame, DWORD timeout)
{
  int            rxCount;
  fd_set         readset;
  struct timeval sockTimeout;

  sockTimeout.tv_sec  = timeout / 1000;
  sockTimeout.tv_usec = (timeout % 1000) * 1000;
  FD_ZERO(&readset);
  FD_SET(pCh->fd, &readset);
  rxCount = select(pCh->fd + 1, &readset, NULL, NULL, &sockTimeout);
  if (rxCount > 0) {
    rxCount = recv(pCh->fd, frame, count, 0);
    if (rxCount !=  -1) return rxCount;
    else                return BDI_SOCKET_ERROR;
  } /* if */
  else if (rxCount == 0) {
    return BDI_SOCKET_RX_TIMEOUT;
  } /* else if */
  else {
    return BDI_SOCKET_ERROR;
  } /* else */
} /* NetWaitFrame */


/****************************************************************************
    Reset Link

     INPUT:  -
     OUTPUT: return         error code

 ****************************************************************************/

static int NetResetLink(BDI_ChannelT* pCh)
{
  int   result;
  int   rxCount;
  int   txCount;
  int   repeats;
  BYTE* txPtr;

  txPtr    = pCh->txFrame;
  *txPtr++ = FRAME_LNK_TYPE;
  *txPtr++ = 1;
  *txPtr++ = LNK_RESET;
  txCount  = txPtr - pCh->txFrame;
  repeats = 0;
  while (repeats++ < 6) {
    rxCount = 0;
    result  = NetSendFrame(pCh, txCount, pCh->txFrame);
    if (result == BDI_OKAY) rxCount = NetWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 500);
    if (rxCount == txCount) return BDI_OKAY;
  } /* while */
  return BDI_ERR_NO_RESPONSE;
} /* NetResetLink */


/****************************************************************************
 ****************************************************************************
                Common Communication Functions
 ****************************************************************************
 ****************************************************************************/


/****************************************************************************
 ****************************************************************************

    BDI_Open:

     Opens the connection to the BDI.

     INPUT  : host      a string for the comm port "/dev/..." or IP number
              param     the baudrate for serial connection
                        the port number for UDP/IP connection
     OUTPUT : RETURN    0 if okay or a negativ number if error.

 ****************************************************************************/

int BDI_Open(BDI_ChannelT* pCh, const char *host, DWORD param)
{
  int   result;
  int   i;

  /* check if already connected */
  result = BDI_OKAY;
  if (pCh->connected) {
    if (pCh->asynConnection) AsynClose(pCh);
    else                     NetClose(pCh);
    pCh->connected = FALSE;
  } /* if */

  /* asyn channel */
  if (strncmp(host, "/dev", 4) == 0) {
    pCh->asynConnection = TRUE;
    result = AsynOpen(pCh, host);
    if (result == BDI_OKAY) {
      i = 0;
      do {
        result = AsynResetLink(pCh, param);
        i++;
      } while ((result != BDI_OKAY) && (i < 3));
      if (result != BDI_OKAY) {
        AsynClose(pCh);
      } /* if */
    } /* if */
  } /* if */

  /* network channel */
  else {
    pCh->asynConnection = FALSE;
    result = NetOpen(pCh, host, (WORD)param);
    if (result == BDI_OKAY) {
      result = NetResetLink(pCh);
      if (result != BDI_OKAY) {
        NetClose(pCh);
      } /* if */
    } /* if */
  } /* else if */

  if (result == BDI_OKAY) {
    pCh->connected   = TRUE;
    pCh->lastError   = BDI_OKAY;
    pCh->frameCount  = 0;
    pCh->repeatCount = 0;
    pCh->bdiEvents   = 0;
    pCh->frameType   = FRAME_STD_TYPE;
  } /* if */

  return result;
} /* BDI_Open */


/****************************************************************************
 ****************************************************************************

    BDI_Close:

     Closes the connection to the BDI.

     INPUT  : -
     OUTPUT : RETURN        0 if okay or a negativ number if error.

 ****************************************************************************/

int BDI_Close(BDI_ChannelT* pCh)
{
  if (pCh->connected) {
    if (pCh->asynConnection) AsynClose(pCh);
    else                     NetClose(pCh);
    pCh->connected = FALSE;
  } /* if */
  return BDI_OKAY;
} /* BDI_Close */


/****************************************************************************
 ****************************************************************************

    BDI_Transaction:

     Executes a command / answer transaction with the BDI
     commandTime: The time the command needs to execute, not the transfer time.
                  The time is used to calculate the timeout before the command
                  is repeated.

     INPUT  : commandLength the length of the command block
              commandData   the command <code,parameter>
              answerSize    the size of the answer buffer
              commandTime   the time in ms the command needs to execute
     OUTPUT : answerData    the answer <code,parameter>
              RETURN        the length of the answer block
                            or a negativ number if error.

 ****************************************************************************/

int   BDI_Transaction(      BDI_ChannelT* pCh,
                            int           commandLength,
                      const void*         commandData,
                            int           answerSize,
                            void*         answerData,
                            DWORD         commandTime)
{
        BYTE*   framePtr;
  const BYTE*   commandPtr;
        BYTE*   answerPtr;
        BYTE    frameControl;
        int     txFrameLength;
        int     rxFrameLength;
        int     sendCount;
        int     result;
        int     rxCount;
        DWORD   answerTimeout;
        BOOL    sendFrame;

  /* check if everthing is okay */
  result = BDI_OKAY;
  if (!pCh->connected) result = BDI_ERR_NOT_CONNECTED;
  else if (pCh->lastError != BDI_OKAY) result = pCh->lastError;
  else if (commandLength > (int)(sizeof pCh->txFrame - 2)) result = BDI_ERR_INVALID_PARAMETER;
  if (result != BDI_OKAY) {
    return result;
  } /* if */

  /* build frame */
  txFrameLength = commandLength + 2;
  framePtr      = pCh->txFrame;
  frameControl  = (BYTE)(pCh->frameType | (commandLength>>8));
  frameControl |= (pCh->frameCount<<6);
  pCh->frameCount++;

  commandPtr  = (const BYTE*)commandData;
  *framePtr++ = frameControl;
  *framePtr++ = (BYTE)commandLength;
  while (commandLength--) *framePtr++ = *commandPtr++;

  /* do transaction */
  sendCount = 0;
  sendFrame = TRUE;
  rxCount   = 0;
  do {

    /* send command frame */
    if (sendFrame) {
      if (pCh->asynConnection) {
        result = AsynSendFrame(pCh, txFrameLength, pCh->txFrame);
      } /* if */
      else {
        result = NetSendFrame(pCh, txFrameLength, pCh->txFrame);
      } /* else */
      sendCount++;
    } /* if */

    /* wait for answer */
    if (result == BDI_OKAY) {
      if (pCh->asynConnection) {
        answerTimeout = (DWORD)(txFrameLength + 1500);  /* total number of characters */
        answerTimeout = answerTimeout * 10000L / pCh->asynBaudrate + commandTime + 200L;
        rxFrameLength = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, answerTimeout);
      } /* if */
      else {
        answerTimeout = NET_TRANSFER_TIMEOUT + commandTime;
        rxFrameLength = NetWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, answerTimeout);
      } /* else */

      /* check if attention frame received */
      if ((rxFrameLength == 3) && (pCh->rxFrame[0] == FRAME_ATT_TYPE) &&  (pCh->rxFrame[1] == 1)) {
        pCh->bdiEvents |= pCh->rxFrame[2];
        sendFrame = !sendFrame;  /* recover from lost command frame */
      } /* if */

      /* check received frame */
      else if (    (rxFrameLength > 2)
                && ((frameControl & FRAME_COUNT_FIELD) == (pCh->rxFrame[0] & FRAME_COUNT_FIELD))
              ) {
        rxFrameLength -= 2;
        rxCount = 256 * (pCh->rxFrame[0] & FRAME_LENGTH_MASK) + pCh->rxFrame[1];
        if (rxFrameLength == rxCount) {
          if (rxCount <= answerSize) {
            framePtr  = pCh->rxFrame + 2;
            answerPtr = (BYTE*)answerData;
            while (rxFrameLength--) *answerPtr++ = *framePtr++;
          } /* if */
        } /* if */
        else {
          rxCount = 0;
          pCh->repeatCount++;
          sendFrame = TRUE;
        } /* else */
      } /* if */

      /* discard wrong frame */
      else if (rxFrameLength > 0) {
        sendFrame = FALSE;
      } /* else if */

      else {
        pCh->repeatCount++;
        commandTime *= 2;  /* increase timeout */
        sendFrame = TRUE;
      } /* if */

    } /* if */
  } while ((rxCount == 0) && (sendCount < 5));

  if (rxCount > answerSize) {
    return BDI_ERR_ANSWER_TOO_BIG;
  } /* if */
  else if (rxCount > 0) {
    return rxCount;
  } /* else if */
  else {
    if (pCh->frameType == FRAME_STD_TYPE) pCh->lastError = BDI_ERR_NO_RESPONSE;
    return BDI_ERR_NO_RESPONSE;
  } /* else */
} /* BDI_Transaction */


/****************************************************************************
 ****************************************************************************

    BDI_GetPendingEvents:

     Returns pending target events. The events are bit coded in a byte:

     bit 0 = target state has changed
     bit 1 = terminal characters are ready to read

     After a call to this function, the events are cleared (only local).
     The BDI will send pending events again after a time interval.

     INPUT  : -
     OUTPUT : RETURN        bit code events.

 ****************************************************************************/

BYTE BDI_LnkGetPendingEvents(BDI_ChannelT* pCh)
{
  BYTE  events;
  int   rxFrameLength = 0;

  /* check if connected */
  if (!pCh->connected) {
    return 0;
  } /* if */

  /* try to get a communication frame */
  if (pCh->asynConnection) {
    if (AsynInRxQueue(pCh) > 0)
      rxFrameLength = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 100);
  } /* if */
  else {
    rxFrameLength = NetWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, 0);
  } /* else */

  /* check if attention frame received */
  if (rxFrameLength == 3) {
    if ((pCh->rxFrame[0] == FRAME_ATT_TYPE) &&  (pCh->rxFrame[1] == 1)) {
      pCh->bdiEvents |= pCh->rxFrame[2];
    } /* if */
  } /* if */

  /* return pending events */
  events = pCh->bdiEvents;
  pCh->bdiEvents = 0;
  return events;
} /* BDI_LnkGetPendingEvents */



/****************************************************************************
 ****************************************************************************
                Fast Download Function
 ****************************************************************************
 ****************************************************************************/

/****************************************************************************
 ****************************************************************************

    BDI_DownloadBlock:

     Download one frame.

     INPUT  : address       the address where to load the block
              count         the number of bytes in the block (max. 1024)
              data          the data block
     OUTPUT : RETURN        0 if okay or a negativ number if error.

 ****************************************************************************/

static int BDI_DownloadBlock(BDI_ChannelT* pCh, DWORD address, WORD count, BYTE* data)
{
  BYTE*   framePtr;
  BYTE    frameControl;
  BYTE    loadResult;
  int     txFrameLength;
  int     rxFrameLength;
  int     repeats;
  int     result;
  int     nextIndex;
  DWORD   repSeqNumber;

  /* wait for unacknowledged frames */
  repeats = 0;
  while ((pCh->reqSeqNumber - pCh->ackSeqNumber) >= pCh->loadWindow) {
    if (pCh->asynConnection)
      rxFrameLength = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, pCh->loadRepeatTime);
    else
      rxFrameLength = NetWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, pCh->loadRepeatTime);

    /* check received frame */
    if (rxFrameLength > 0) {
      if (pCh->rxFrame[2] == BDI_CMD_LOAD_BLOCK) {
        framePtr   = pCh->rxFrame + 3;
        loadResult = *framePtr++;
        framePtr   = BDI_ExtractLong(&pCh->ackSeqNumber,  framePtr);
        if (loadResult > 0) {
          framePtr = BDI_ExtractLong(&pCh->loadErrorAddr, framePtr);
          return BDI_ERR_MEM_ACCESS;
        } /* if */
      } /* if */
    } /* if */

    /* repeat sending of frames */
    else {
      if (repeats++ > 3) return BDI_ERR_NO_RESPONSE;
      for (repSeqNumber = pCh->ackSeqNumber; repSeqNumber < pCh->reqSeqNumber; repSeqNumber++) {
        nextIndex = (int)(repSeqNumber % LOAD_FRAME_COUNT);
        if (pCh->asynConnection)
          result = AsynSendFrame(pCh, pCh->downloadCount[nextIndex], pCh->downloadFrame[nextIndex]);
        else
          result = NetSendFrame(pCh, pCh->downloadCount[nextIndex], pCh->downloadFrame[nextIndex]);
      } /* for */
      pCh->repeatCount++;
    } /* else */
  } /* while */


  /* send next frame */
  frameControl  = (BYTE)(pCh->frameType | (pCh->frameCount<<6));
  pCh->frameCount++;

  nextIndex   = (int)(pCh->reqSeqNumber % LOAD_FRAME_COUNT);
  framePtr    = pCh->downloadFrame[nextIndex];
  framePtr   += 2; /* room for frame control and length */
  *framePtr++ = BDI_CMD_LOAD_BLOCK;
  *framePtr++ = pCh->compressType;
  framePtr    = BDI_AppendLong(pCh->reqSeqNumber, framePtr);
  framePtr    = BDI_AppendLong(address,      framePtr);
  framePtr    = BDI_AppendWord(count,        framePtr);

  /* append data part */
  if (pCh->compressType > 0) {
    framePtr += LZA_Compress(pCh->hCompress, count, data, framePtr);
  } /* if */
  else {
    while (count--) *framePtr++ = *data++;
  } /* else */

  /* calculate and update frame control/length */
  txFrameLength = framePtr - pCh->downloadFrame[nextIndex];
  framePtr      = pCh->downloadFrame[nextIndex];
  frameControl |= ((txFrameLength-2) >> 8);
  *framePtr++   = frameControl;
  *framePtr++   = (BYTE)(txFrameLength-2);

  pCh->downloadCount[nextIndex] = txFrameLength;
  pCh->reqSeqNumber++;

  if (pCh->asynConnection)
    result = AsynSendFrame(pCh, txFrameLength, pCh->downloadFrame[nextIndex]);
  else
    result = NetSendFrame(pCh, txFrameLength, pCh->downloadFrame[nextIndex]);

  return result;
} /* BDI_DownloadBlock */


/****************************************************************************
 ****************************************************************************

    BDI_TerminateDownload:

     Terminates the download via bdiNet.
     Wait for all answers and repeats frames if neccessary.

     INPUT  : -
     OUTPUT : RETURN        0 if okay or a negativ number if error.

 ****************************************************************************/

static int BDI_TerminateDownload(BDI_ChannelT* pCh)
{
  BYTE*   framePtr;
  BYTE    loadResult;
  int     rxFrameLength;
  int     repeats;
  int     result;
  int     nextIndex;
  DWORD   repSeqNumber;

  /* wait for unacknowledged frames */
  repeats = 0;
  while (pCh->reqSeqNumber > pCh->ackSeqNumber) {
    if (pCh->asynConnection)
      rxFrameLength = AsynWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, pCh->loadRepeatTime);
    else
      rxFrameLength = NetWaitFrame(pCh, sizeof pCh->rxFrame, pCh->rxFrame, pCh->loadRepeatTime);

    /* check received frame */
    if (rxFrameLength > 0) {
      if (pCh->rxFrame[2] == BDI_CMD_LOAD_BLOCK) {
        framePtr   = pCh->rxFrame + 3;
        loadResult = *framePtr++;
        framePtr   = BDI_ExtractLong(&pCh->ackSeqNumber,  framePtr);
        if (loadResult > 0) {
          framePtr   = BDI_ExtractLong(&pCh->loadErrorAddr, framePtr);
          return BDI_ERR_MEM_ACCESS;
        } /* if */
      } /* if */
    } /* if */

    /* repeat sending of outstanding frames */
    else {
      if (repeats++ > 3) return BDI_ERR_NO_RESPONSE;
      for (repSeqNumber = pCh->ackSeqNumber; repSeqNumber < pCh->reqSeqNumber; repSeqNumber++) {
        nextIndex = (int)(repSeqNumber % LOAD_FRAME_COUNT);
        if (pCh->asynConnection)
          result = AsynSendFrame(pCh, pCh->downloadCount[nextIndex], pCh->downloadFrame[nextIndex]);
        else
          result = NetSendFrame(pCh, pCh->downloadCount[nextIndex], pCh->downloadFrame[nextIndex]);
      } /* for */
      pCh->repeatCount++;
    } /* else */
  } /* while */
  return BDI_OKAY;
} /* BDI_TerminateDownload */


/****************************************************************************
 ****************************************************************************

    BDI_InitDownload:

     Initilisation for the fast code download.
	 If successful, the channel is locked by the semaphore until
	 BDI_EndDownload() is called.

     INPUT  : verifyMode    defines how to verify the written data
                            0 = no verify at all (the fastest mode)
                            1 = verify only first byte of block
                            2 = reads back all data from the target memory
              compress      If set to TRUE, data compression is used
     OUTPUT : RETURN        0 if okay or a negativ number if error.

 ****************************************************************************/

int BDI_LnkInitDownload(BDI_ChannelT* pCh, BYTE verifyMode, BOOL compress)
{
  BYTE* framePtr;
  int   rxCount;
  BYTE  cmdBuffer[2];
  BYTE  ansBuffer[6];

  pCh->loadError     = BDI_OKAY;
  pCh->loadErrorAddr = 0;
  pCh->reqSeqNumber  = 0;
  pCh->ackSeqNumber  = 0;
  if (pCh->asynConnection) pCh->compressType = (BYTE)(compress ? 1 : 0);
  else                     pCh->compressType = 0;

  framePtr = BDI_AppendByte(BDI_CMD_LOAD_INIT, cmdBuffer);
  framePtr = BDI_AppendByte(verifyMode, framePtr);
  rxCount  = BDI_Transaction(pCh, framePtr-cmdBuffer, cmdBuffer, sizeof ansBuffer, ansBuffer, 10);
  if ((rxCount == sizeof ansBuffer) && (ansBuffer[0] == BDI_CMD_LOAD_INIT)) {
    framePtr = BDI_ExtractByte(&pCh->loadWindow, ansBuffer+1);
    framePtr = BDI_ExtractWord(&pCh->loadMaxDataSize, framePtr);
    framePtr = BDI_ExtractWord(&pCh->loadRepeatTime,  framePtr);
    if (pCh->loadWindow > 3) pCh->loadWindow  = 3;
    if (pCh->loadMaxDataSize > LOAD_MAX_BLOCK_SIZE) pCh->loadMaxDataSize = LOAD_MAX_BLOCK_SIZE;
    if (pCh->asynConnection) pCh->loadRepeatTime += 2000;
    else                     pCh->loadRepeatTime += 100;
    if (pCh->compressType > 0) {
      pCh->hCompress = LZA_CompressOpen();
    } /* if */
    return BDI_OKAY;
  } /* if */
  else if (rxCount < 0) return rxCount;
  else                  return BDI_ERR_INVALID_RESPONSE;
} /* BDI_LnkInitDownload */


/****************************************************************************
 ****************************************************************************

    BDI_Download:

     Download a block of data. If an error is returned, get the cause with
     BDI_EndDownload.

     INPUT  : address       the address where to load the block
              count         the number of bytes in the block (no upper limit)
              data          the data block
     OUTPUT : RETURN        0 if okay or a negativ number if error.

 ****************************************************************************/

int BDI_LnkDownload(BDI_ChannelT* pCh, DWORD address, WORD count, BYTE* data)
{
  WORD  currentCount;

  while ((count > 0) && (pCh->loadError == BDI_OKAY)) {
    /* get next partial block */
    if (count > pCh->loadMaxDataSize) currentCount = pCh->loadMaxDataSize;
    else                              currentCount = count;
    pCh->loadError = BDI_DownloadBlock(pCh, address, currentCount, data);
    count    -= currentCount;
    address  += currentCount;
    data     += currentCount;
  } /* while */

  return pCh->loadError;
} /* BDI_LnkDownload */


/****************************************************************************
 ****************************************************************************

    BDI_EndDownload:

     Terminates the fast download process. The result of the download
     is returned.

     INPUT  : -
     OUTPUT : errorAddr     the address where the first error was detected
              RETURN        0 if okay or a negativ number if error.

 ****************************************************************************/

int BDI_LnkEndDownload(BDI_ChannelT* pCh, DWORD* errorAddr)
{
  if (pCh->loadError == BDI_OKAY) {
    pCh->loadError = BDI_TerminateDownload(pCh);
  } /* if */
  *errorAddr = pCh->loadErrorAddr;
  if (pCh->compressType > 0) {
    LZA_CompressClose(pCh->hCompress);
  } /* if */
  return pCh->loadError;
} /* BDI_LnkEndDownload */



